@@ -1725,8 +1725,8 @@ static u64 get_symbol_size(Symbol<E> &sym) {
 }
 
 template <typename E>
-ElfSym<E> to_output_esym(Context<E> &ctx, Symbol<E> &sym, u32 st_name,
-                         U32<E> *shn_xindex) {
+std::optional<ElfSym<E>>
+to_output_esym(Context<E> &ctx, Symbol<E> &sym, u32 st_name, U32<E> *shn_xindex) {
   ElfSym<E> esym;
   memset(&esym, 0, sizeof(esym));
 
@@ -1835,7 +1835,8 @@ ElfSym<E> to_output_esym(Context<E> &ctx, Symbol<E> &sym, u32 st_name,
   if (0 <= shndx && shndx < SHN_LORESERVE) {
     esym.st_shndx = shndx;
   } else if (SHN_LORESERVE <= shndx) {
-    assert(shn_xindex);
+    if (!shn_xindex)
+      return {};
     esym.st_shndx = SHN_XINDEX;
     *shn_xindex = shndx;
   }
@@ -1869,7 +1870,17 @@ void DynsymSection<E>::copy_buf(Context<E> &ctx) {
 
   for (i64 i = 1; i < symbols.size(); i++) {
     Symbol<E> &sym = *symbols[i];
-    buf[sym.get_dynsym_idx(ctx)] = to_output_esym(ctx, sym, offset, nullptr);
+
+    std::optional<ElfSym<E>> esym = to_output_esym(ctx, sym, offset, nullptr);
+    if (!esym) {
+      Error(ctx) << ctx.arg.output
+                 << ": .dynsym: too many output sections: "
+                 << (ctx.shdr->shdr.sh_size / sizeof(ElfShdr<E>))
+                 << " requested, but ELF allows at most 65279";
+      return;
+    }
+
+    buf[sym.get_dynsym_idx(ctx)] = *esym;
     offset += sym.name().size() + 1;
   }
 }
@@ -1981,7 +1992,7 @@ get_merged_output_name(Context<E> &ctx, std::string_view name, u64 flags,
                        i64 entsize, i64 addralign) {
   if (ctx.arg.relocatable && !ctx.arg.relocatable_merge_sections)
     return name;
-  if (ctx.arg.unique && ctx.arg.unique->match(name))
+  if (!ctx.arg.unique.empty() && ctx.arg.unique.find(name) != -1)
     return name;
 
   // GCC seems to create sections named ".rodata.strN.<mangled-symbol-name>.M"
@@ -2827,42 +2838,33 @@ void NotePropertySection<E>::copy_buf(Context<E> &ctx) {
 
 template <typename E>
 CompressedSection<E>::CompressedSection(Context<E> &ctx, Chunk<E> &chunk) {
-  assert(chunk.name.starts_with(".debug"));
-  this->name = chunk.name;
-  this->is_compressed = true;
+  // Allocate a temporary buffer to write uncompressed contents. Note
+  // that we use u8[] instead of std::vector<u8> to avoid the cost of
+  // zero-initialization, as sh_size can be very large.
+  std::unique_ptr<u8[]> buf(new u8[chunk.shdr.sh_size]);
 
-  this->uncompressed_data.resize(chunk.shdr.sh_size);
-  u8 *buf = this->uncompressed_data.data();
-
-  chunk.write_to(ctx, buf);
-
-  switch (ctx.arg.compress_debug_sections) {
-  case COMPRESS_ZLIB:
-    chdr.ch_type = ELFCOMPRESS_ZLIB;
-    compressor.reset(new ZlibCompressor(buf, chunk.shdr.sh_size));
-    break;
-  case COMPRESS_ZSTD:
-    chdr.ch_type = ELFCOMPRESS_ZSTD;
-    compressor.reset(new ZstdCompressor(buf, chunk.shdr.sh_size));
-    break;
-  default:
-    unreachable();
-  }
+  // Write uncompressed contents and then compress them
+  chunk.write_to(ctx, buf.get());
+  compressor.emplace(ctx.arg.compress_debug_sections, buf.get(),
+                     chunk.shdr.sh_size);
 
+  // Compute header field values
+  chdr.ch_type = ctx.arg.compress_debug_sections;
   chdr.ch_size = chunk.shdr.sh_size;
   chdr.ch_addralign = chunk.shdr.sh_addralign;
 
+  this->name = chunk.name;
+  this->shndx = chunk.shndx;
+  this->is_compressed = true;
+
   this->shdr = chunk.shdr;
   this->shdr.sh_flags |= SHF_COMPRESSED;
   this->shdr.sh_addralign = 1;
   this->shdr.sh_size = sizeof(chdr) + compressor->compressed_size;
-  this->shndx = chunk.shndx;
 
-  // We don't need to keep the original data unless --gdb-index is given.
-  if (!ctx.arg.gdb_index) {
-    this->uncompressed_data.clear();
-    this->uncompressed_data.shrink_to_fit();
-  }
+  // We can discard the uncompressed contents unless --gdb-index is given
+  if (ctx.arg.gdb_index)
+    this->uncompressed_data = std::move(buf);
 }
 
 template <typename E>
@@ -3045,6 +3047,8 @@ template class GnuDebuglinkSection<E>;
 template Chunk<E> *find_chunk(Context<E> &, u32);
 template Chunk<E> *find_chunk(Context<E> &, std::string_view);
 template i64 to_phdr_flags(Context<E> &ctx, Chunk<E> *chunk);
-template ElfSym<E> to_output_esym(Context<E> &, Symbol<E> &, u32, U32<E> *);
+
+template std::optional<ElfSym<E>>
+to_output_esym(Context<E> &, Symbol<E> &, u32, U32<E> *);
 
 } // namespace mold
